/*
 * livepatch_bsc1201742
 *
 * Fix for CVE-2020-36557, bsc#1201742 and CVE-2020-36558, bsc#1201752
 *
 *  Upstream commits:
 *  ca4463bf8438 ("vt: vt_ioctl: fix VT_DISALLOCATE freeing in-use virtual
 *                 console")
 *  6cd1ed50efd8 ("vt: vt_ioctl: fix race in VT_RESIZEX")
 *
 *  SLE12-SP4, SLE12-SP5, SLE15-SP1 commits:
 *  e6f55bd64169f4b281743ad1346178d9b470cf98
 *  664d504c1361a64bc6a2f777b60ce1888067ec62
 *
 *  SLE15 commits:
 *  f15e18db1c9e0d107b5e4c51a731a7c70858e9b7
 *  3c76a1f3602a29e6dc6d9de6ea4398bdac789036
 *
 *  SLE15-SP2 and -SP3 commits:
 *  3c829773522850768eabd7ad0596b58e0265a6fc
 *  6f9e172b2a161598e6258039f69a3c1fbe0a8e26
 *
 *  SLE15-SP4 commits:
 *  not affected
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_VT)

/* klp-ccp: from drivers/tty/vt/vt_ioctl.c */
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched/signal.h>
#include <linux/tty.h>

/* klp-ccp: from include/linux/tty.h */
int klpp_vt_ioctl(struct tty_struct *tty,
		    unsigned int cmd, unsigned long arg);

/* klp-ccp: from drivers/tty/vt/vt_ioctl.c */
#include <linux/timer.h>
#include <linux/kernel.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/console.h>
#include <linux/signal.h>
#include <linux/timex.h>
#include <asm/io.h>
#include <linux/uaccess.h>
#include <linux/nospec.h>

/* klp-ccp: from include/linux/kbd_kern.h */
static int (*klpe_set_console)(int nr);

/* klp-ccp: from drivers/tty/vt/vt_ioctl.c */
#include <linux/vt_kern.h>

/* klp-ccp: from include/linux/vt_kern.h */
static int (*klpe_kbd_rate)(struct kbd_repeat *rep);

static int (*klpe_vc_allocate)(unsigned int console);
static int (*klpe_vc_cons_allocated)(unsigned int console);

static struct vc_data *(*klpe_vc_deallocate)(unsigned int console);

static int (*klpe_con_font_op)(struct vc_data *vc, struct console_font_op *op);
static int (*klpe_con_set_cmap)(unsigned char __user *cmap);
static int (*klpe_con_get_cmap)(unsigned char __user *cmap);

static int (*klpe_tioclinux)(struct tty_struct *tty, unsigned long arg);

#ifdef CONFIG_CONSOLE_TRANSLATIONS

static int (*klpe_con_set_trans_old)(unsigned char __user * table);
static int (*klpe_con_get_trans_old)(unsigned char __user * table);
static int (*klpe_con_set_trans_new)(unsigned short __user * table);
static int (*klpe_con_get_trans_new)(unsigned short __user * table);
static int (*klpe_con_clear_unimap)(struct vc_data *vc);
static int (*klpe_con_set_unimap)(struct vc_data *vc, ushort ct, struct unipair __user *list);
static int (*klpe_con_get_unimap)(struct vc_data *vc, ushort ct, ushort __user *uct, struct unipair __user *list);

#else
#error "klp-ccp: non-taken branch"
#endif

static int (*klpe_vt_waitactive)(int n);

static char (*klpe_vt_dont_switch);

static struct vt_spawn_console (*klpe_vt_spawn_con);

static int (*klpe_vt_do_diacrit)(unsigned int cmd, void __user *up, int eperm);
static int (*klpe_vt_do_kdskbmode)(int console, unsigned int arg);
static int (*klpe_vt_do_kdskbmeta)(int console, unsigned int arg);
static int (*klpe_vt_do_kbkeycode_ioctl)(int cmd, struct kbkeycode __user *user_kbkc,
								int perm);
static int (*klpe_vt_do_kdsk_ioctl)(int cmd, struct kbentry __user *user_kbe,
					int perm, int console);
static int (*klpe_vt_do_kdgkb_ioctl)(int cmd, struct kbsentry __user *user_kdgkb,
                                        int perm);
static int (*klpe_vt_do_kdskled)(int console, int cmd, unsigned long arg, int perm);
static int (*klpe_vt_do_kdgkbmode)(int console);
static int (*klpe_vt_do_kdgkbmeta)(int console);

/* klp-ccp: from include/linux/selection.h */
static struct vc_data *(*klpe_sel_cons);

/* klp-ccp: from drivers/tty/vt/vt_ioctl.c */
static struct tty_driver *(*klpe_console_driver);

#define KLPR_VT_IS_IN_USE(i)	((*klpe_console_driver)->ttys[i] && (*klpe_console_driver)->ttys[i]->count)
#define KLPR_VT_BUSY(i)	(KLPR_VT_IS_IN_USE(i) || i == fg_console || vc_cons[i].d == (*klpe_sel_cons))


#ifdef CONFIG_X86

/* klp-ccp: from include/linux/syscalls.h */
static long (*klpe_sys_ioperm)(unsigned long from, unsigned long num, int on);

/* klp-ccp: from drivers/tty/vt/vt_ioctl.c */
#endif

static void (*klpe_complete_change_console)(struct vc_data *vc);

static int (*klpe_vt_event_wait_ioctl)(struct vt_event __user *event);

#define GPFIRST 0x3b4
#define GPLAST 0x3df
#define GPNUM (GPLAST - GPFIRST + 1)

static inline int 
klpr_do_fontx_ioctl(struct vc_data *vc, int cmd, struct consolefontdesc __user *user_cfd, int perm, struct console_font_op *op)
{
	struct consolefontdesc cfdarg;
	int i;

	if (copy_from_user(&cfdarg, user_cfd, sizeof(struct consolefontdesc))) 
		return -EFAULT;
 	
	switch (cmd) {
	case PIO_FONTX:
		if (!perm)
			return -EPERM;
		op->op = KD_FONT_OP_SET;
		op->flags = KD_FONT_FLAG_OLD;
		op->width = 8;
		op->height = cfdarg.charheight;
		op->charcount = cfdarg.charcount;
		op->data = cfdarg.chardata;
		return (*klpe_con_font_op)(vc, op);

	case GIO_FONTX:
		op->op = KD_FONT_OP_GET;
		op->flags = KD_FONT_FLAG_OLD;
		op->width = 8;
		op->height = cfdarg.charheight;
		op->charcount = cfdarg.charcount;
		op->data = cfdarg.chardata;
		i = (*klpe_con_font_op)(vc, op);
		if (i)
			return i;
		cfdarg.charheight = op->height;
		cfdarg.charcount = op->charcount;
		if (copy_to_user(user_cfd, &cfdarg, sizeof(struct consolefontdesc)))
			return -EFAULT;
		return 0;
	}
	return -EINVAL;
}

static inline int 
klpr_do_unimap_ioctl(int cmd, struct unimapdesc __user *user_ud, int perm, struct vc_data *vc)
{
	struct unimapdesc tmp;

	if (copy_from_user(&tmp, user_ud, sizeof tmp))
		return -EFAULT;
	if (tmp.entries)
		if (!access_ok(VERIFY_WRITE, tmp.entries,
				tmp.entry_ct*sizeof(struct unipair)))
			return -EFAULT;
	switch (cmd) {
	case PIO_UNIMAP:
		if (!perm)
			return -EPERM;
		return (*klpe_con_set_unimap)(vc, tmp.entry_ct, tmp.entries);
	case GIO_UNIMAP:
		if (!perm && fg_console != vc->vc_num)
			return -EPERM;
		return (*klpe_con_get_unimap)(vc, tmp.entry_ct, &(user_ud->entry_ct), tmp.entries);
	}
	return 0;
}

/* New. */
static bool klpp_is_vc_in_use(unsigned int vc_num) {
	struct vc_data *vc;

	if (!(*klpe_vc_cons_allocated)(vc_num))
		return false;

	vc = vc_cons[vc_num].d;

	return vc && vc->port.itty != NULL;
}

static int klpp_vt_disallocate(unsigned int vc_num)
{
	struct vc_data *vc = NULL;
	int ret = 0;

	console_lock();
	if (KLPR_VT_BUSY(vc_num))
		ret = -EBUSY;
	/*
	 * Fix CVE-2020-36557
	 *  +2 lines
	 */
	else if (klpp_is_vc_in_use(vc_num))
		ret = -EBUSY;
	else if (vc_num)
		vc = (*klpe_vc_deallocate)(vc_num);
	console_unlock();

	if (vc && vc_num >= MIN_NR_CONSOLES) {
		tty_port_destroy(&vc->port);
		kfree(vc);
	}

	return ret;
}

static void klpp_vt_disallocate_all(void)
{
	struct vc_data *vc[MAX_NR_CONSOLES];
	int i;

	console_lock();
	for (i = 1; i < MAX_NR_CONSOLES; i++)
		/*
		 * Fix CVE-2020-36557
		 *  -1 line, +1 line
		 */
		if (!KLPR_VT_BUSY(i) && !klpp_is_vc_in_use(i))
			vc[i] = (*klpe_vc_deallocate)(i);
		else
			vc[i] = NULL;
	console_unlock();

	for (i = 1; i < MAX_NR_CONSOLES; i++) {
		if (vc[i] && i >= MIN_NR_CONSOLES) {
			tty_port_destroy(&vc[i]->port);
			kfree(vc[i]);
		}
	}
}

int klpp_vt_ioctl(struct tty_struct *tty,
	     unsigned int cmd, unsigned long arg)
{
	struct vc_data *vc = tty->driver_data;
	struct console_font_op op;	/* used in multiple places here */
	unsigned int console;
	unsigned char ucval;
	unsigned int uival;
	void __user *up = (void __user *)arg;
	int i, perm;
	int ret = 0;

	console = vc->vc_num;


	if (!(*klpe_vc_cons_allocated)(console)) { 	/* impossible? */
		ret = -ENOIOCTLCMD;
		goto out;
	}


	/*
	 * To have permissions to do most of the vt ioctls, we either have
	 * to be the owner of the tty, or have CAP_SYS_TTY_CONFIG.
	 */
	perm = 0;
	if (current->signal->tty == tty || capable(CAP_SYS_TTY_CONFIG))
		perm = 1;
 
	switch (cmd) {
	case TIOCLINUX:
		ret = (*klpe_tioclinux)(tty, arg);
		break;
	case KIOCSOUND:
		if (!perm)
			return -EPERM;
		/*
		 * The use of PIT_TICK_RATE is historic, it used to be
		 * the platform-dependent CLOCK_TICK_RATE between 2.6.12
		 * and 2.6.36, which was a minor but unfortunate ABI
		 * change. kd_mksound is locked by the input layer.
		 */
		if (arg)
			arg = PIT_TICK_RATE / arg;
		kd_mksound(arg, 0);
		break;

	case KDMKTONE:
		if (!perm)
			return -EPERM;
	{
		unsigned int ticks, count;
		
		/*
		 * Generate the tone for the appropriate number of ticks.
		 * If the time is zero, turn off sound ourselves.
		 */
		ticks = msecs_to_jiffies((arg >> 16) & 0xffff);
		count = ticks ? (arg & 0xffff) : 0;
		if (count)
			count = PIT_TICK_RATE / count;
		kd_mksound(count, ticks);
		break;
	}

	case KDGKBTYPE:
		/*
		 * this is naïve.
		 */
		ucval = KB_101;
		ret = put_user(ucval, (char __user *)arg);
		break;

#ifdef CONFIG_X86
	case KDADDIO:
	case KDDELIO:
		/*
		 * KDADDIO and KDDELIO may be able to add ports beyond what
		 * we reject here, but to be safe...
		 *
		 * These are locked internally via sys_ioperm
		 */
		if (arg < GPFIRST || arg > GPLAST) {
			ret = -EINVAL;
			break;
		}
		ret = (*klpe_sys_ioperm)(arg, 1, (cmd == KDADDIO)) ? -ENXIO : 0;
		break;

	case KDENABIO:
	case KDDISABIO:
		ret = (*klpe_sys_ioperm)(GPFIRST, GPNUM,
				  (cmd == KDENABIO)) ? -ENXIO : 0;
		break;
#endif
	case KDKBDREP:
	{
		struct kbd_repeat kbrep;
		
		if (!capable(CAP_SYS_TTY_CONFIG))
			return -EPERM;

		if (copy_from_user(&kbrep, up, sizeof(struct kbd_repeat))) {
			ret =  -EFAULT;
			break;
		}
		ret = (*klpe_kbd_rate)(&kbrep);
		if (ret)
			break;
		if (copy_to_user(up, &kbrep, sizeof(struct kbd_repeat)))
			ret = -EFAULT;
		break;
	}

	case KDSETMODE:
		/*
		 * currently, setting the mode from KD_TEXT to KD_GRAPHICS
		 * doesn't do a whole lot. i'm not sure if it should do any
		 * restoration of modes or what...
		 *
		 * XXX It should at least call into the driver, fbdev's definitely
		 * need to restore their engine state. --BenH
		 */
		if (!perm)
			return -EPERM;
		switch (arg) {
		case KD_GRAPHICS:
			break;
		case KD_TEXT0:
		case KD_TEXT1:
			arg = KD_TEXT;
		case KD_TEXT:
			break;
		default:
			ret = -EINVAL;
			goto out;
		}
		console_lock();
		if (vc->vc_mode == (unsigned char) arg) {
			console_unlock();
			break;
		}
		vc->vc_mode = (unsigned char) arg;
		if (console != fg_console) {
			console_unlock();
			break;
		}
		/*
		 * explicitly blank/unblank the screen if switching modes
		 */
		if (arg == KD_TEXT)
			do_unblank_screen(1);
		else
			do_blank_screen(1);
		console_unlock();
		break;

	case KDGETMODE:
		uival = vc->vc_mode;
		goto setint;

	case KDMAPDISP:
	case KDUNMAPDISP:
		/*
		 * these work like a combination of mmap and KDENABIO.
		 * this could be easily finished.
		 */
		ret = -EINVAL;
		break;

	case KDSKBMODE:
		if (!perm)
			return -EPERM;
		ret = (*klpe_vt_do_kdskbmode)(console, arg);
		if (ret == 0)
			tty_ldisc_flush(tty);
		break;

	case KDGKBMODE:
		uival = (*klpe_vt_do_kdgkbmode)(console);
		ret = put_user(uival, (int __user *)arg);
		break;

	/* this could be folded into KDSKBMODE, but for compatibility
	   reasons it is not so easy to fold KDGKBMETA into KDGKBMODE */
	case KDSKBMETA:
		ret = (*klpe_vt_do_kdskbmeta)(console, arg);
		break;

	case KDGKBMETA:
		/* FIXME: should review whether this is worth locking */
		uival = (*klpe_vt_do_kdgkbmeta)(console);
	setint:
		ret = put_user(uival, (int __user *)arg);
		break;

	case KDGETKEYCODE:
	case KDSETKEYCODE:
		if(!capable(CAP_SYS_TTY_CONFIG))
			perm = 0;
		ret = (*klpe_vt_do_kbkeycode_ioctl)(cmd, up, perm);
		break;

	case KDGKBENT:
	case KDSKBENT:
		ret = (*klpe_vt_do_kdsk_ioctl)(cmd, up, perm, console);
		break;

	case KDGKBSENT:
	case KDSKBSENT:
		ret = (*klpe_vt_do_kdgkb_ioctl)(cmd, up, perm);
		break;

	/* Diacritical processing. Handled in keyboard.c as it has
	   to operate on the keyboard locks and structures */
	case KDGKBDIACR:
	case KDGKBDIACRUC:
	case KDSKBDIACR:
	case KDSKBDIACRUC:
		ret = (*klpe_vt_do_diacrit)(cmd, up, perm);
		break;

	/* the ioctls below read/set the flags usually shown in the leds */
	/* don't use them - they will go away without warning */
	case KDGKBLED:
	case KDSKBLED:
	case KDGETLED:
	case KDSETLED:
		ret = (*klpe_vt_do_kdskled)(console, cmd, arg, perm);
		break;

	/*
	 * A process can indicate its willingness to accept signals
	 * generated by pressing an appropriate key combination.
	 * Thus, one can have a daemon that e.g. spawns a new console
	 * upon a keypress and then changes to it.
	 * See also the kbrequest field of inittab(5).
	 */
	case KDSIGACCEPT:
	{
		if (!perm || !capable(CAP_KILL))
			return -EPERM;
		if (!valid_signal(arg) || arg < 1 || arg == SIGKILL)
			ret = -EINVAL;
		else {
			spin_lock_irq(&(*klpe_vt_spawn_con).lock);
			put_pid((*klpe_vt_spawn_con).pid);
			(*klpe_vt_spawn_con).pid = get_pid(task_pid(current));
			(*klpe_vt_spawn_con).sig = arg;
			spin_unlock_irq(&(*klpe_vt_spawn_con).lock);
		}
		break;
	}

	case VT_SETMODE:
	{
		struct vt_mode tmp;

		if (!perm)
			return -EPERM;
		if (copy_from_user(&tmp, up, sizeof(struct vt_mode))) {
			ret = -EFAULT;
			goto out;
		}
		if (tmp.mode != VT_AUTO && tmp.mode != VT_PROCESS) {
			ret = -EINVAL;
			goto out;
		}
		console_lock();
		vc->vt_mode = tmp;
		/* the frsig is ignored, so we set it to 0 */
		vc->vt_mode.frsig = 0;
		put_pid(vc->vt_pid);
		vc->vt_pid = get_pid(task_pid(current));
		/* no switch is required -- saw@shade.msu.ru */
		vc->vt_newvt = -1;
		console_unlock();
		break;
	}

	case VT_GETMODE:
	{
		struct vt_mode tmp;
		int rc;

		console_lock();
		memcpy(&tmp, &vc->vt_mode, sizeof(struct vt_mode));
		console_unlock();

		rc = copy_to_user(up, &tmp, sizeof(struct vt_mode));
		if (rc)
			ret = -EFAULT;
		break;
	}

	/*
	 * Returns global vt state. Note that VT 0 is always open, since
	 * it's an alias for the current VT, and people can't use it here.
	 * We cannot return state for more than 16 VTs, since v_state is short.
	 */
	case VT_GETSTATE:
	{
		struct vt_stat __user *vtstat = up;
		unsigned short state, mask;

		/* Review: FIXME: Console lock ? */
		if (put_user(fg_console + 1, &vtstat->v_active))
			ret = -EFAULT;
		else {
			state = 1;	/* /dev/tty0 is always open */
			for (i = 0, mask = 2; i < MAX_NR_CONSOLES && mask;
							++i, mask <<= 1)
				if (((*klpe_console_driver)->ttys[i] && (*klpe_console_driver)->ttys[i]->count))
					state |= mask;
			ret = put_user(state, &vtstat->v_state);
		}
		break;
	}

	/*
	 * Returns the first available (non-opened) console.
	 */
	case VT_OPENQRY:
		/* FIXME: locking ? - but then this is a stupid API */
		for (i = 0; i < MAX_NR_CONSOLES; ++i)
			if (! ((*klpe_console_driver)->ttys[i] && (*klpe_console_driver)->ttys[i]->count))
				break;
		uival = i < MAX_NR_CONSOLES ? (i+1) : -1;
		goto setint;		 

	/*
	 * ioctl(fd, VT_ACTIVATE, num) will cause us to switch to vt # num,
	 * with num >= 1 (switches to vt 0, our console, are not allowed, just
	 * to preserve sanity).
	 */
	case VT_ACTIVATE:
		if (!perm)
			return -EPERM;
		if (arg == 0 || arg > MAX_NR_CONSOLES)
			ret =  -ENXIO;
		else {
			arg--;
			console_lock();
			ret = (*klpe_vc_allocate)(arg);
			console_unlock();
			if (ret)
				break;
			(*klpe_set_console)(arg);
		}
		break;

	case VT_SETACTIVATE:
	{
		struct vt_setactivate vsa;

		if (!perm)
			return -EPERM;

		if (copy_from_user(&vsa, (struct vt_setactivate __user *)arg,
					sizeof(struct vt_setactivate))) {
			ret = -EFAULT;
			goto out;
		}
		if (vsa.console == 0 || vsa.console > MAX_NR_CONSOLES)
			ret = -ENXIO;
		else {
			vsa.console = array_index_nospec(vsa.console,
							 MAX_NR_CONSOLES + 1);
			vsa.console--;
			console_lock();
			ret = (*klpe_vc_allocate)(vsa.console);
			if (ret == 0) {
				struct vc_data *nvc;
				/* This is safe providing we don't drop the
				   console sem between vc_allocate and
				   finishing referencing nvc */
				nvc = vc_cons[vsa.console].d;
				nvc->vt_mode = vsa.mode;
				nvc->vt_mode.frsig = 0;
				put_pid(nvc->vt_pid);
				nvc->vt_pid = get_pid(task_pid(current));
			}
			console_unlock();
			if (ret)
				break;
			/* Commence switch and lock */
			/* Review set_console locks */
			(*klpe_set_console)(vsa.console);
		}
		break;
	}

	/*
	 * wait until the specified VT has been activated
	 */
	case VT_WAITACTIVE:
		if (!perm)
			return -EPERM;
		if (arg == 0 || arg > MAX_NR_CONSOLES)
			ret = -ENXIO;
		else
			ret = (*klpe_vt_waitactive)(arg);
		break;

	/*
	 * If a vt is under process control, the kernel will not switch to it
	 * immediately, but postpone the operation until the process calls this
	 * ioctl, allowing the switch to complete.
	 *
	 * According to the X sources this is the behavior:
	 *	0:	pending switch-from not OK
	 *	1:	pending switch-from OK
	 *	2:	completed switch-to OK
	 */
	case VT_RELDISP:
		if (!perm)
			return -EPERM;

		console_lock();
		if (vc->vt_mode.mode != VT_PROCESS) {
			console_unlock();
			ret = -EINVAL;
			break;
		}
		/*
		 * Switching-from response
		 */
		if (vc->vt_newvt >= 0) {
			if (arg == 0)
				/*
				 * Switch disallowed, so forget we were trying
				 * to do it.
				 */
				vc->vt_newvt = -1;

			else {
				/*
				 * The current vt has been released, so
				 * complete the switch.
				 */
				int newvt;
				newvt = vc->vt_newvt;
				vc->vt_newvt = -1;
				ret = (*klpe_vc_allocate)(newvt);
				if (ret) {
					console_unlock();
					break;
				}
				/*
				 * When we actually do the console switch,
				 * make sure we are atomic with respect to
				 * other console switches..
				 */
				(*klpe_complete_change_console)(vc_cons[newvt].d);
			}
		} else {
			/*
			 * Switched-to response
			 */
			/*
			 * If it's just an ACK, ignore it
			 */
			if (arg != VT_ACKACQ)
				ret = -EINVAL;
		}
		console_unlock();
		break;

	 /*
	  * Disallocate memory associated to VT (but leave VT1)
	  */
	 case VT_DISALLOCATE:
		if (arg > MAX_NR_CONSOLES) {
			ret = -ENXIO;
			break;
		}
		if (arg == 0)
			klpp_vt_disallocate_all();
		else
			ret = klpp_vt_disallocate(--arg);
		break;

	case VT_RESIZE:
	{
		struct vt_sizes __user *vtsizes = up;
		struct vc_data *vc;

		ushort ll,cc;
		if (!perm)
			return -EPERM;
		if (get_user(ll, &vtsizes->v_rows) ||
		    get_user(cc, &vtsizes->v_cols))
			ret = -EFAULT;
		else {
			console_lock();
			for (i = 0; i < MAX_NR_CONSOLES; i++) {
				vc = vc_cons[i].d;

				if (vc) {
					vc->vc_resize_user = 1;
					/* FIXME: review v tty lock */
					vc_resize(vc_cons[i].d, cc, ll);
				}
			}
			console_unlock();
		}
		break;
	}

	case VT_RESIZEX:
	{
		struct vt_consize __user *vtconsize = up;
		ushort ll,cc,vlin,clin,vcol,ccol;
		if (!perm)
			return -EPERM;
		if (!access_ok(VERIFY_READ, vtconsize,
				sizeof(struct vt_consize))) {
			ret = -EFAULT;
			break;
		}
		/* FIXME: Should check the copies properly */
		__get_user(ll, &vtconsize->v_rows);
		__get_user(cc, &vtconsize->v_cols);
		__get_user(vlin, &vtconsize->v_vlin);
		__get_user(clin, &vtconsize->v_clin);
		__get_user(vcol, &vtconsize->v_vcol);
		__get_user(ccol, &vtconsize->v_ccol);
		vlin = vlin ? vlin : vc->vc_scan_lines;
		if (clin) {
			if (ll) {
				if (ll != vlin/clin) {
					/* Parameters don't add up */
					ret = -EINVAL;
					break;
				}
			} else 
				ll = vlin/clin;
		}
		if (vcol && ccol) {
			if (cc) {
				if (cc != vcol/ccol) {
					ret = -EINVAL;
					break;
				}
			} else
				cc = vcol/ccol;
		}

		if (clin > 32) {
			ret =  -EINVAL;
			break;
		}
		    
		for (i = 0; i < MAX_NR_CONSOLES; i++) {
			/*
			 * Fix CVE-2020-36558
			 *  +2 lines
			 */
			struct vc_data *vcp;

			if (!vc_cons[i].d)
				continue;
			console_lock();
			/*
			 * Fix CVE-2020-36558
			 *  -6 lines, +9lines
			 */
			vcp = vc_cons[i].d;
			if (vcp) {
				if (vlin)
					vcp->vc_scan_lines = vlin;
				if (clin)
					vcp->vc_font.height = clin;
				vcp->vc_resize_user = 1;
				vc_resize(vcp, cc, ll);
			}
			console_unlock();
		}
		break;
	}

	case PIO_FONT: {
		if (!perm)
			return -EPERM;
		op.op = KD_FONT_OP_SET;
		op.flags = KD_FONT_FLAG_OLD | KD_FONT_FLAG_DONT_RECALC;	/* Compatibility */
		op.width = 8;
		op.height = 0;
		op.charcount = 256;
		op.data = up;
		ret = (*klpe_con_font_op)(vc, &op);
		break;
	}

	case GIO_FONT: {
		op.op = KD_FONT_OP_GET;
		op.flags = KD_FONT_FLAG_OLD;
		op.width = 8;
		op.height = 32;
		op.charcount = 256;
		op.data = up;
		ret = (*klpe_con_font_op)(vc, &op);
		break;
	}

	case PIO_CMAP:
                if (!perm)
			ret = -EPERM;
		else
	                ret = (*klpe_con_set_cmap)(up);
		break;

	case GIO_CMAP:
                ret = (*klpe_con_get_cmap)(up);
		break;

	case PIO_FONTX:
	case GIO_FONTX:
		ret = klpr_do_fontx_ioctl(vc, cmd, up, perm, &op);
		break;

	case PIO_FONTRESET:
	{
		if (!perm)
			return -EPERM;

#ifdef BROKEN_GRAPHICS_PROGRAMS
		ret = -ENOSYS;
		break;
#else
		{
		op.op = KD_FONT_OP_SET_DEFAULT;
		op.data = NULL;
		ret = (*klpe_con_font_op)(vc, &op);
		if (ret)
			break;
		console_lock();
		con_set_default_unimap(vc);
		console_unlock();
		break;
		}
#endif
	}

	case KDFONTOP: {
		if (copy_from_user(&op, up, sizeof(op))) {
			ret = -EFAULT;
			break;
		}
		if (!perm && op.op != KD_FONT_OP_GET)
			return -EPERM;
		ret = (*klpe_con_font_op)(vc, &op);
		if (ret)
			break;
		if (copy_to_user(up, &op, sizeof(op)))
			ret = -EFAULT;
		break;
	}

	case PIO_SCRNMAP:
		if (!perm)
			ret = -EPERM;
		else
			ret = (*klpe_con_set_trans_old)(up);
		break;

	case GIO_SCRNMAP:
		ret = (*klpe_con_get_trans_old)(up);
		break;

	case PIO_UNISCRNMAP:
		if (!perm)
			ret = -EPERM;
		else
			ret = (*klpe_con_set_trans_new)(up);
		break;

	case GIO_UNISCRNMAP:
		ret = (*klpe_con_get_trans_new)(up);
		break;

	case PIO_UNIMAPCLR:
		if (!perm)
			return -EPERM;
		(*klpe_con_clear_unimap)(vc);
		break;

	case PIO_UNIMAP:
	case GIO_UNIMAP:
		ret = klpr_do_unimap_ioctl(cmd, up, perm, vc);
		break;

	case VT_LOCKSWITCH:
		if (!capable(CAP_SYS_TTY_CONFIG))
			return -EPERM;
		(*klpe_vt_dont_switch) = 1;
		break;
	case VT_UNLOCKSWITCH:
		if (!capable(CAP_SYS_TTY_CONFIG))
			return -EPERM;
		(*klpe_vt_dont_switch) = 0;
		break;
	case VT_GETHIFONTMASK:
		ret = put_user(vc->vc_hi_font_mask,
					(unsigned short __user *)arg);
		break;
	case VT_WAITEVENT:
		ret = (*klpe_vt_event_wait_ioctl)((struct vt_event __user *)arg);
		break;
	default:
		ret = -ENOIOCTLCMD;
	}
out:
	return ret;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1201742.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "complete_change_console", (void *)&klpe_complete_change_console },
	{ "con_clear_unimap", (void *)&klpe_con_clear_unimap },
	{ "con_font_op", (void *)&klpe_con_font_op },
	{ "con_get_cmap", (void *)&klpe_con_get_cmap },
	{ "con_get_trans_new", (void *)&klpe_con_get_trans_new },
	{ "con_get_trans_old", (void *)&klpe_con_get_trans_old },
	{ "con_get_unimap", (void *)&klpe_con_get_unimap },
	{ "con_set_cmap", (void *)&klpe_con_set_cmap },
	{ "con_set_trans_new", (void *)&klpe_con_set_trans_new },
	{ "con_set_trans_old", (void *)&klpe_con_set_trans_old },
	{ "con_set_unimap", (void *)&klpe_con_set_unimap },
	{ "console_driver", (void *)&klpe_console_driver },
	{ "kbd_rate", (void *)&klpe_kbd_rate },
	{ "sel_cons", (void *)&klpe_sel_cons },
	{ "set_console", (void *)&klpe_set_console },
#ifdef CONFIG_X86
	{ "sys_ioperm", (void *)&klpe_sys_ioperm },
#endif
	{ "tioclinux", (void *)&klpe_tioclinux },
	{ "vc_allocate", (void *)&klpe_vc_allocate },
	{ "vc_cons_allocated", (void *)&klpe_vc_cons_allocated },
	{ "vc_deallocate", (void *)&klpe_vc_deallocate },
	{ "vt_do_diacrit", (void *)&klpe_vt_do_diacrit },
	{ "vt_do_kbkeycode_ioctl", (void *)&klpe_vt_do_kbkeycode_ioctl },
	{ "vt_do_kdgkb_ioctl", (void *)&klpe_vt_do_kdgkb_ioctl },
	{ "vt_do_kdgkbmeta", (void *)&klpe_vt_do_kdgkbmeta },
	{ "vt_do_kdgkbmode", (void *)&klpe_vt_do_kdgkbmode },
	{ "vt_do_kdsk_ioctl", (void *)&klpe_vt_do_kdsk_ioctl },
	{ "vt_do_kdskbmeta", (void *)&klpe_vt_do_kdskbmeta },
	{ "vt_do_kdskbmode", (void *)&klpe_vt_do_kdskbmode },
	{ "vt_do_kdskled", (void *)&klpe_vt_do_kdskled },
	{ "vt_dont_switch", (void *)&klpe_vt_dont_switch },
	{ "vt_event_wait_ioctl", (void *)&klpe_vt_event_wait_ioctl },
	{ "vt_spawn_con", (void *)&klpe_vt_spawn_con },
	{ "vt_waitactive", (void *)&klpe_vt_waitactive },
};

int livepatch_bsc1201742_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}

#endif /* IS_ENABLED(CONFIG_VT) */
